/*
 * umaxscan.c - driver program for the Umax Astra 1220U USB scanner.
 *
 * Paul Mackerras, December 1999
 *
 * Copyright 1999 Paul Mackerras.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 */

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>

int fd;
int xper;

struct cmsg_ioctl {
	unsigned char	requesttype;
	unsigned char	request;
	unsigned char	value[2];
	unsigned char	index[2];
	unsigned char	length[2];
	void		*data;
};

void control_msg(int rtype, int req, int value, int index, int len,
		 void *data)
{
	struct cmsg_ioctl c;

	c.requesttype = rtype;
	c.request = req;
	c.value[0] = value;
	c.value[1] = value >> 8;
	c.index[0] = index;
	c.index[1] = index >> 8;
	c.length[0] = len;
	c.length[1] = len >> 8;
	c.data = data;
	if (ioctl(fd, 0x556d, &c) < 0) {
		perror("scanner control");
		exit(1);
	}
}

int getstat(int which)
{
	unsigned char buf[1];

	control_msg(0xc0, 0, 0, which, 1, buf);
	return buf[0];
}

void xpectstat(int which, int val)
{
	int s = getstat(which);
	if (s != val) {
		printf("gstat %d -> %x (expected %x)\n", which, s, val);
		exit(1);
	}
}

void waitstat(int which, int val, int timeout)
{
	int s, n = 0;

	for (;;) {
		s = getstat(which);
		if (s == val)
			return;
		if (++n >= timeout) {
			printf("waiting for gstat %d -> %x (got %x)\n",
			       which, val, s);
			exit(1);
		}
		usleep(100000);
	}
}

void xxxops(void)
{
	control_msg(0x40, 1, 2, 3, 0, 0);
	control_msg(0x40, 1, 14, 2, 0, 0);
	control_msg(0x40, 1, 64, 0, 0, 0);
	control_msg(0x40, 1, 6, 2, 0, 0);
	xpectstat(4, 0x38);
	control_msg(0x40, 1, 7, 2, 0, 0);
	xpectstat(4, 0x38);
	control_msg(0x40, 1, 4, 2, 0, 0);
	xpectstat(4, 0xf8);
	control_msg(0x40, 1, 5, 2, 0, 0);
	xpectstat(2, 5);
	control_msg(0x40, 1, 4, 2, 0, 0);
	control_msg(0x40, 1, 30, 3, 0, 0);
}

int verbose;

void send(unsigned char *buf, int nb)
{
	int ret;

	control_msg(0x40, 4, 0, 0, 0, 0);
	control_msg(0x40, 3, nb, 0, 0, 0);
	ret = write(fd, buf, nb);
	if (ret < 0) {
		perror("write");
		exit(1);
	}
}

void usync(int len, int cmd)
{
	unsigned char buf[4];
	int s0, s4;

	buf[0] = 0x55;
	buf[1] = 0xaa;
	send(buf, 2);
	waitstat(4, 0xc8, 20);
	buf[0] = len >> 16;
	buf[1] = len >> 8;
	buf[2] = len;
	buf[3] = cmd;
	send(buf, 4);
	s0 = getstat(0);
	s4 = getstat(4);
}

void qsend(unsigned char *buf, int nb)
{
	int s0, s4, ret;

	waitstat(4, 0xc8, 20);
	send(buf, nb);
	s4 = getstat(4);
	s0 = getstat(0);
}

int qread(unsigned char *buf, int nb)
{
	int s4, s0, ret, n;

	waitstat(4, 0xd0, 2000);
	for (; nb > 0; nb -= n) {
		n = nb;
		if (n > 0xf000)
			n = 0xf000;
		control_msg(0x40, 2, n, 0, 0, 0);
		ret = read(fd, buf, n);
		if (ret < 0) {
			fprintf(stderr, "reading %d(%d): ", n, nb);
			perror(NULL);
			exit(1);
		}
		if (ret < n) {
			fprintf(stderr, "reading %d(%d) only got %d\n",
				n, nb, ret);
			return nb - ret;
		}
		buf += n;
	}
	s4 = getstat(4);
	s0 = getstat(0);
	return nb;
}

unsigned char buf[16384];

int bescape(unsigned char *data, int dlen)
{
	unsigned char *p, *q;
	int i, c;

	p = data;
	q = buf;
	for (i = 0; i < dlen; ++i) {
		c = *p++;
		if (c == 0x1b || (c == 0xaa && i > 0 && p[-2] == 0x55))
			*q++ = 0x1b;
		*q++ = c;
	}
	return q - buf;
}

void cwrite(int len, int cmd, unsigned char *data, int dlen)
{
	usync(len, cmd | 0x80);
	if (dlen <= 0)
		return;
	dlen = bescape(data, dlen);
	qsend(buf, dlen);
}

void cwritev(int len, int cmd, unsigned char *data, int dlen)
{
	cwrite(len, cmd, data, dlen);
	if (dlen <= 0)
		return;
	usync(len, cmd | 0xc0);
	qread(buf, len);
	if (bcmp(buf, data, len)) {
		printf("data mismatch on cmd %x, %x\n", len, cmd);
		exit(1);
	}
}

int cread(int len, int cmd, unsigned char *data)
{
	usync(len, cmd | 0xc0);
	if (len <= 0)
		return 0;
	return qread(data, len);
}

unsigned char opb[36] = {
	0, 0, 4, 0, 2, 0, 0, 12, 0, 3, 0xc1, 0x80, 0x60, 0x20, 0, 0,
	0x16, 0x41, 0xe0, 0xac, 3, 3, 0, 0, 0x46, 0xa0, 0, 0, 0, 0,
	0, 0, 0, 0xf0,
};
unsigned char opb1[36] = {
	0, 0, 4, 0, 2, 0, 0, 12, 0, 3, 0xc1, 0x80, 0, 0x20, 2, 0,
	0x16, 0x41, 0xe0, 0xac, 3, 3, 0, 0, 0x46, 0xa0, 0, 0, 0, 0,
	0, 0, 0, 0x10, 0x1a
};
unsigned char opb2[36] = {
	0, 0, 6, 0, 2, 0, 0, 12, 0, 3, 0xc1, 0x80, 0, 0x20, 2, 0,
	0x16, 0x41, 0xe0, 0xac, 3, 3, 0, 0, 0x46, 0xa0, 0, 0, 0, 0,
	0, 0, 0, 0x10, 0x1a,
};
unsigned char opb3[36] = {
	0, 0, 4, 0, 2, 0, 0, 12, 0, 3, 0xc1, 0x80, 0, 0, 4, 0,
	0x16, 0x80, 0x15, 0x78, 3, 3, 0, 0, 0x46, 0xa0, 0, 0x8b, 0x49, 0x4a,
	0xd0, 0x68, 0xdf, 0x1b, 0x1a,
};
unsigned char opb4[36] = {
	0, 0, 4, 0, 2, 0, 0, 12, 0, 3, 0xc1, 0x80, 0x60, 0x20, 0, 0,
	0x16, 0x41, 0xe0, 0xac, 3, 3, 0, 0, 0x46, 0xa0, 0, 0x8b, 0x49, 0x4a,
	0xd0, 0x68, 0xdf, 0xf3, 0x1b
};
unsigned char opb8[36] = {
	0, 0, 4, 0, 2, 0, 0, 12, 0, 4, 0x40, 1, 0, 0, 4, 0,
	0x6e, 0x18, 0x10, 3, 6, 0, 0, 0, 0x46, 0xa0, 0, 0x8b, 0x49, 0x4a,
	0xd0, 0x68, 0xdf, 0x13, 0x1a
};
unsigned char opb9[36] = {
	0, 0, 4, 0, 2, 0, 0, 12, 0, 4, 0x40, 1, 0, 0, 4, 0,
	0x6e, 0x41, 0x20, 0x24, 6, 0, 0, 0, 0x46, 0xa0, 0, 0x8b, 0x49, 0x4a,
	0xd0, 0x68, 0xdf, 0x13, 0x1a
};
unsigned char opb10[36] = {
	0, 0, 4, 0, 2, 0, 0, 12, 0, 4, 0x40, 1, 0, 0, 4, 0,
	0x6e, 0x41, 0x60, 0x4f, 6, 0, 0, 0, 0x46, 0xa0, 0, 0x8b, 0x49, 0x4a,
	0xd0, 0x68, 0xdf, 0x93, 0x1a
};
unsigned char opbx[36];

unsigned char opc[16] = {
	2, 0x80, 0, 0x70, 0, 0, 0x60, 0x2f, 0x2f, 7, 0, 0, 0, 0x80, 0xf0, 0
};
unsigned char opc1[16] = {
	1, 0, 1, 0x70, 0, 0, 0x60, 0x2f, 0x13, 5, 0, 0, 0, 0x80, 0xf0, 0
};
unsigned char opc5[16] = {
	5, 0, 0, 0x70, 0, 0, 0x60, 0x2f, 0x2f, 0, 0, 0x30, 12, 0xc3, 0xa4, 0
};
unsigned char opc6[16] = {
	1, 0, 0, 0x70, 0, 0, 0x60, 0x2f, 0x2f, 0, 0x88, 0x48, 12, 0x83, 0xa4, 0
};
unsigned char opc7[16] = {
	5, 0, 0, 0x70, 0, 0, 0x60, 0x2f, 0x2f, 0, 0xec, 0x4e, 12, 0xc3, 0xa4, 0
};
unsigned char opcx[16];

unsigned char opd[8] = {
	0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa
};
unsigned char opd2[8] = {
	6, 0xf4, 0xff, 0x81, 0x3d, 0, 0, 0x30
};

unsigned char ope[8] = {
	0, 0, 0, 0, 0, 0, 0xff, 0xff
};
unsigned char ope1[8] = {
	0, 0, 0, 0xaa, 0xcc, 0xee, 0x80, 0xff
};

unsigned char ramp[800];
unsigned char dbuf[65536];

void umaxinit()
{
	int i, s;
	unsigned char *p;

	xxxops();
	usync(0, 0);
	xxxops();

	cwritev(0x22, 8, opb, 0x22);
	cread(0x23, 8, opbx);
	cwritev(0x23, 8, opb1, 0x23);

	usync(0, 0xc2);
	s = getstat(0);

	p = ramp;
	*p++ = 0;
	*p++ = 0;
	*p++ = 0;
	for (i = 0; i < 768; ++i)
		*p++ = i;
	*p++ = 0xaa;
	*p++ = 0xaa;
	cwritev(p - ramp, 4, ramp, p - ramp);

	cwritev(0x23, 8, opb1, 0x23);

	usync(0, 0xc2);
	s = getstat(0);

	p = ramp;
	for (i = 0; i < 256; ++i) {
		*p++ = i;
		*p++ = 0;
	}
	cwrite(p - ramp, 4, ramp, p - ramp);

	cwritev(0x23, 8, opb2, 0x23);

	usync(0, 0xc2);
	s = getstat(0);

	p = ramp;
	for (i = 0; i < 256; ++i) {
		*p++ = i;
		*p++ = 4;
	}
	cwritev(p - ramp, 4, ramp, p - ramp);

	cwritev(0x23, 8, opb1, 0x23);
	cwritev(0x10, 2, opc, 0x10);
	cwritev(8, 1, opd, 8);

	usync(0, 0);
	usync(0, 0xc2);
	s = getstat(0);
	if ((s & 0x40) == 0) {
		cwritev(0x10, 2, opc1, 0x10);
		cwritev(0x23, 8, opb3, 0x23);

		usync(0, 0x40);

		for (i = 0; i < 60; ++i) {
			usync(0, 0xc2);
			s = getstat(0);
			if ((s & 0x40) != 0)
				break;
			usleep(500000);
		}
	}

	usync(0, 0);
}

void umaxinit2(void)
{
	int s;

	xxxops();
	usync(0, 0);
	xxxops();

	cwritev(0x22, 8, opb4, 0x23);
	cread(0x23, 8, opbx);
	cread(0x10, 2, opcx);
	cwritev(0x10, 2, opc, 0x10);
	cwritev(8, 1, opd, 8);

	usync(0, 0);
	usync(0, 0xc2);
	s = getstat(0);
}

unsigned char opc4[16] = {
	1, 0, 0, 0x20, 0, 0, 0x60, 0x2f, 0x2f, 1, 0, 0, 0, 0x80, 0xa4, 0
};
unsigned char opb5[36] = {
	0, 0, 4, 0, 2, 0, 0, 12, 0, 4, 0x40, 1, 0, 0, 4, 0,
	0x6e, 0xf6, 0x79, 0xbf, 6, 0, 0, 0, 0x46, 0xa0, 0, 0x8b, 0x49, 0x4a,
	0xd0, 0x68, 0xdf, 0x13, 0x1a
};
unsigned char opb7[36] = {
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0,
	0x6e, 0xf6, 0x79, 0xbf, 1, 0, 0, 0, 0x46, 0xa0, 0, 0x8b, 0x49, 0x4a,
	0xd0, 0x68, 0xdf, 0x13, 0x1a
};
unsigned char ope2[3] = {
	0, 0xff, 0x8f
};

int scanner_ypos = 0;
int scanner_yorg;

void move(int distance, int fine)
{
	int s;
	int rev = distance < 0;
	int skip = (rev? -distance: distance) - 1;

	if (distance == 0)
		return;
	opc4[1] = skip << 6;
	opc4[2] = skip >> 2;
	opc4[3] = (rev? 0x20: 0x70) + ((skip >> 10) & 0xf);
	opc4[9] = rev? 1: 5;
	scanner_ypos += (fine? distance: 2 * distance + (rev? -1: 1));
	scanner_ypos = (scanner_ypos + (rev? 0: 3)) & ~3;
	if (fine) {
		opc4[8] = 0x2f;
		opc4[14] = 0xa4;
	} else {
		opc4[8] = 0x17;
		opc4[14] = 0xac;
	}
	cwrite(0x10, 2, opc4, 0x10);
	cwrite(0x23, 8, rev? opb7: opb5, 0x23);

	usync(0, 0xc2);
	s = getstat(0);
	usync(0, 0);
	if (rev)
		cwrite(3, 4, ope2, 3);
	else
		cwrite(8, 4, ope, 8);

	usync(0, 0x40);

	usync(0, 0xc2);
	s = getstat(0);
	usync(0, 0xc2);
	s = getstat(0);

	cread(0x200, 4, dbuf);
}

void get_pixels(unsigned char *op2, unsigned char *op8, unsigned char *op1,
		unsigned char *op4, int len, int zpos, unsigned char *buf)
{
	int s;

	if (zpos == 0)
		usync(0, 0);
	cwrite(0x10, 2, op2, 0x10);
	cwrite(0x23, 8, op8, 0x23);
	cwrite(8, 1, op1, 8);
	usync(0, 0xc2);
	s = getstat(0);
	if (zpos == 1)
		usync(0, 0);
	cwrite(8, 4, op4, 8);
	usync(0, 0x40);
	usync(0, 0xc2);
	s = getstat(0);
	usync(0, 0xc2);
	s = getstat(0);
	cread(len, 4, (buf? buf: dbuf));
}

int locate_black_stripe(unsigned char *img, int w, int h)
{
	int epos, ecnt, x, y;
	unsigned char *p;

	epos = 0;
	ecnt = 0;
	p = img;
	for (x = 0; x < w; ++x, ++p) {
		int d, dmax=0, dpos=0;
		unsigned char *q = img + x;
		for (y = 1; y < h; ++y, q += w) {
			d = q[0] - q[w];
			if (d > dmax) {
				dmax = d;
				dpos = y;
			}
		}
		if (dmax > 0) {
			epos += dpos;
			++ecnt;
		}
	}
	if (ecnt == 0)
		epos = 70;
	else
		epos = (epos + ecnt/2) / ecnt;
	return epos;
}

unsigned char opc3[16] = {
	0xb4, 0, 0, 0x70, 0, 0, 0x60, 0x2f, 0x2f, 5, 0, 0, 0, 0x80, 0xa4, 0
};
unsigned char opb6[36] = {
	0, 0, 4, 0, 2, 0, 0, 12, 0, 4, 0x40, 1, 0, 0, 4, 0,
	0x6e, 0xfb, 0xc4, 0xe5, 6, 0, 0, 0x60, 0x4d, 0xa0, 0, 0x8b, 0x49, 0x4a,
	0xd0, 0x68, 0xdf, 0x13, 0x1a
};
unsigned char opd1[8] = {
	6, 0xf4, 0xff, 0x81, 0x3d, 0, 8, 0
};

void find_zero(void)
{
	int s;
	unsigned char *img;
	FILE *f;

	img = malloc(54000);
	if (img == 0) {
		fprintf(stderr, "out of memory (need 54000)\n");
		exit(1);
	}

	usync(0, 0);
	get_pixels(opc3, opb6, opd1, ope1, 54000, 1, img);

	s = locate_black_stripe(img, 300, 180);
	scanner_yorg = scanner_ypos + s + 64;
	scanner_ypos += 180 + 3;
	scanner_ypos &= ~3;

	free(img);
}

#if 0
void calib(void)
{
	opc5[11] = 0x30;
	opd2[7] = 0x30;
	get_pixels(opc5, opb8, opd2, ope, 24, 0, 0);

	opc5[11] = 0x40;
	opd2[7] = 0x40;
	get_pixels(opc5, opb8, opd2, ope, 24, 0, 0);

	opd2[6] = 8;
	opd2[7] = 0x30;
	get_pixels(opc6, opb9, opd2, ope, 0x200, 1, 0);

	opc7[10] = 0xec;
	opd2[6] = 0xc;
	opd2[7] = 0x40;
	get_pixels(opc7, opb10, opd2, ope, 5300, 1, 0);

	opc7[10] = 0xed;
	opd2[6] = 0xd;
	get_pixels(opc7, opb10, opd2, ope, 5300, 0, 0);
}
#endif

void restore(void)
{
	usync(0, 0);
	cwritev(0x10, 2, opc1, 0x10);
	cwritev(0x23, 8, opb3, 0x23);
	usync(0, 0x40);
	scanner_ypos = 0;
}

void lamp(int on)
{
	opc1[14] = on? 0xf0: 0x90;
	usync(0, 0);
	xxxops();
	usync(0, 0);
	cwrite(16, 2, opc1, 16);
}

/* reorder BGR lines to RGB pixels */
void reorder_rgb(unsigned char *p, int roff, int goff, int boff,
		 int w, int h, int pitch, FILE *f)
{
	int x, y;

	for (y = 0; y < h; ++y) {
		for (x = 0; x < w; ++x, ++p) {
			putc(p[roff], f);
			putc(p[goff], f);
			putc(p[boff], f);
		}
		p += pitch - w;
	}
}

unsigned char caldata[16070];

unsigned char opc9[16] = {
	0, 0, 0, 0x70, 0, 0, 0x60, 0, 0x17, 5, 0xec, 0x4e, 12, 0xc3, 0xac
};
unsigned char opb11[36] = {
	0, 0, 4, 0, 2, 0, 0, 12, 0, 4, 0x40, 1, 0, 0, 4, 0,
	0x6e, 0xad, 0xa0, 0x49, 6, 0, 0, 0xec, 0x54, 0xa0, 0, 0x8b, 0x49, 0x4a,
	0xd0, 0x68, 0xdf, 0x93, 0x1b
};
unsigned char opd4[8] = {
	6, 0xf4, 0xff, 0x81, 0x3d, 0, 0xc, 0x40
};

void get_caldata(int color)
{
	unsigned char *p;
	int h = 66;
	int w = color? 3 * 5100: 5100;
	int x0 = color? 0: 5100;
	int l = w * h;
	int i, x, y;
	FILE *f;

	p = malloc(l);
	if (p == 0) {
		fprintf(stderr, "out of memory (need %d)\n", l);
		exit(1);
	}
	memset(caldata, 0, 3*5100);

	usync(0, 0);
	opc9[0] = h + 4;
	if (color) {
		opc9[13] = 3;
		opb11[23] = 0xc4;
		opb11[24] = 0x5c;
		opd4[6] = 8;
		opd4[7] = 0;
	} else {
		opc9[13] = 0xc3;
		opb11[23] = 0xec;
		opb11[24] = 0x54;
		opd4[6] = 0xc;
		opd4[7] = 0x40;
	}
	get_pixels(opc9, opb11, opd4, ope, l, 0, p);
	scanner_ypos += (h + 4) * 2 + 3;
	scanner_ypos &= ~3;

	for (x = 0; x < w; ++x) {
		int t = 0, gn;
		double av, gain;

		for (y = 0; y < h; ++y)
			t += p[x + y * w];
		av = (double) t / h;
		gain = 250 / av;
		gn = (int) ((gain - 0.984) * 102.547 + 0.5);
		if (gn < 0)
			gn = 0;
		else if (gn > 255)
			gn = 255;
		caldata[x + x0] = gn;
	}
	for (i = 0; i < 768; ++i)
		caldata[i + 3*5100] = i;
	free(p);
}

void dump(unsigned char *buf, int nb)
{
	int i;

	for (i = 0; i < nb; ++i) {
		printf(" %.2x", buf[i]);
		if ((i & 15) == 15)
			printf("\n");
	}
	if ((nb & 15) != 0)
		printf("\n");
}

unsigned char opcgo[16] = {
	0, 0, 0, 0x50, 0xec, 3, 0x60, 0x40, 0x17, 5, 0xec, 0x4e, 12, 0xc3,
};
unsigned char opbgo[36] = {
	0, 0, 0xb0, 0x4f, 0xd8, 0xe7, 0xfa, 16, 0xef, 0xc4, 0x3c, 0x71, 15, 0,
	4, 0, 0x6e, 0, 0, 0, 0xc4, 0x7e, 0, 0xc6, 0x4a, 0xa0, 10, 0x8b,
	0x49, 0x4a, 0xd0, 0x68, 0xdf, 0xb3, 0x1a
};
unsigned char opdgo[8] = {
	6, 0xf4, 0xff, 0x81, 0x3d, 0, 0x8c, 0x40
};

unsigned char subsamp[9] = {
	0xff, 0xff, 0xaa, 0xaa, 0x88, 0x88, 0x88, 0x88, 0x80
};

int red_yoffset = 6;
int green_yoffset = 3;
int blue_yoffset = 0;
int max_yoffset = 6;

void doscan(char *fname, FILE *f, int color, int w, int h,
	    int xsamp, int ysamp, int xskip, int yskip, int ydpi)
{
	int xend = xskip + w * xsamp + (xsamp+1)/2;
	int yoff_scale = 600 * ysamp / ydpi;
	int roff = (8 / yoff_scale * 3 + 2) * w;
	int goff = (4 / yoff_scale * 3 + 1) * w;
	int hextra = color? 8 / yoff_scale: 0;
	int hexp = h + hextra + (color? 4: 0);
	int ytot = hexp * ysamp + 12;
	int bh, bhx, l, s, y;
	unsigned char *p, *q;
	int linelen = color? 3 * w: w;

	bh = 0x80000 / linelen;
	if (bh > hexp)
		bh = hexp;
	l = bh * linelen;
	p = malloc(l);
	if (p == 0) {
		fprintf(stderr, "out of memory (%d)\n", l);
		exit(1);
	}

	opbgo[17] = xskip;
	opbgo[18] = ((xskip >> 8) & 0xf) + (xend << 4);
	opbgo[19] = xend >> 4;
	opbgo[33] = 0x33 + ((xend & 0x1000) >> 5) + ((xskip & 0x1000) >> 6);
	opbgo[23] = color? 0xc6: 0x77;
	opbgo[24] = color? 0x5b: 0x4a;
	opcgo[0] = ytot;
	opcgo[1] = ((ytot >> 8) & 0x3f) + (yskip << 6);
	opcgo[2] = yskip >> 2;
	opcgo[3] = 0x50 + ((yskip >> 10) & 0xf);
	opcgo[6] = bh == h? 0: 0x60;	/* a guess */
	opcgo[7] = color? 47: 64;
	opcgo[8] = ((ydpi <= 300)? 24: 48) - 1;
	opcgo[9] = (ydpi >= 300)? 5: 7;
	opcgo[12] = color? 16: 12;
	opcgo[13] = color? 4: 0xc3;
	opcgo[14] = (ydpi == 600)? 0xa4: 0xac;
	opdgo[6] = color? 0x88: 0x8c;
	opdgo[7] = color? 0: 0x40;
	if (xper) {
		static char line[80];
		int a, v;
		dump(opcgo, 16);
		dump(opbgo, 35);
		dump(opdgo, 8);
		for (;;) {
			printf(": ");
			fflush(stdout);
			if (fgets(line, sizeof(line), stdin) == NULL)
				break;
			if (line[0] == '\n')
				break;
			if (line[0] == '?') {
				dump(opcgo, 16);
				dump(opbgo, 35);
				dump(opdgo, 8);
			} else if (sscanf(line, "%x %x", &a, &v) == 2) {
				if (a < 16)
					opcgo[a] = v;
				else if (a < 64)
					opbgo[a-16] = v;
				else if (a < 72)
					opdgo[a-64] = v;
			}
		}
	}
	usync(0, 0);
	usync(0, 0);
	cwritev(0x10, 2, opcgo, 0x10);
	cwritev(0x23, 8, opbgo, 0x23);
	cwritev(8, 1, opdgo, 8);
	usync(0, 0xc2);
	s = getstat(0);

	caldata[16068] = subsamp[xsamp];
	caldata[16069] = subsamp[ysamp];
	cwrite(16070, 4, caldata, 16070);

	usync(0, 0x40);
	usync(0, 0xc2);
	s = getstat(0);

	fprintf(f, "P%d\n%d %d\n255\n", 5+color, w, h);
	q = p;
	bhx = hextra;
	while (hexp > 0) {
		if (bh > hexp)
			bh = hexp;
		hexp -= bh;
		usync(0, 0xc2);
		s = getstat(0);
		s = cread(bh * linelen, 4, q);
		if (s) {
			fprintf(stderr, "warning: short read (%d bytes)\n", s);
			h = bh;
		}
		if (!color) {
			fwrite(p, 1, bh * linelen, f);
		} else {
			if (bhx != 0) {
				if (xper) {
					FILE *f2 = fopen("strip.pgm", "w");
					fprintf(f2, "P5\n%d %d\n255\n", w, 3*bh);
					fwrite(p, 1, bh*linelen, f2);
					fclose(f2);
				}

				bh -= bhx;
				q += bhx * linelen;
				bhx = 0;
				reorder_rgb(p + 4 * linelen, roff, goff, 0,
					    w, bh - 4, linelen, f);
			} else
				reorder_rgb(p, roff, goff, 0,
					    w, bh, linelen, f);
			memmove(p, p + bh * linelen, hextra * linelen);
		}
	}
	free(p);
}

void parse_2dopt(char *arg, int *xp, int *yp, int opt, int yequalsx)
{
	char *p, *endp;

	*xp = strtol(arg, &endp, 0);
	if (endp == arg)
		goto bad;
	p = endp;
	if (*p == 'x' || *p == 'X' || *p == ',' || *p == '/') {
		++p;
		*yp = strtol(p, &endp, 0);
		if (endp == p || *endp != 0)
			goto bad;
	} else if (yequalsx)
		*yp = *xp;
	return;

 bad:
	fprintf(stderr, "bad argument '%s' for -%c option\n", arg, opt);
}

void usage(const char *av0, int stat)
{
	fprintf(stderr,
"Usage: %s [-c] [-r XxY] [-d XxY] [-o XxY] [-p|-P] [-L|-C|-R] [-O] [file]\n"
"  -c       Color scan (default is greyscale)\n"
"  -r XxY   Set the output image size to X x Y pixels\n"
"  -d XxY   Set the pixel density to X x Y pixels/inch (75, 150, 300, 600)\n"
"           (default is 300, Y same as X if omitted)\n"
"  -o XxY   Offset the image by X x Y units of 1/600 inch\n"
"  -p       Preview: equivalent to -r 640x880 -d 75 -C\n"
"  -P       Preview relative: -r and -o values relative to preview image\n"
"  -L       Image starts at left edge of scanner bed (right, looking down)\n"
"  -C       Image is centered on scanner bed (default)\n"
"  -R       Image ends at right edge of scanner bed (left, looking down)\n"
"  -O       Turn lamp off after scan\n"
"  -h       Print this message\n"
"At least one of -r, -p, -O or -h must be specified\n",
		av0);
	exit(stat);
}

#define ISPOW2(x)	((x) != 0 && ((x) & ((x) - 1)) == 0)

main(int ac, char **av)
{
	FILE *f;
	int s, i;
	unsigned char *p;
	char *outn, *endp;
	int w = 600, h = 400;
	int xo = 0, yo = 0;
	int xres = 300, yres = 300;
	int ydpi;
	int xs, ys, yd;
	int align = 1;	/* default to center */
	int color = 0;
	int prerel = 0;
	int lampoff = 0;
	int res_given = 0;

	while ((i = getopt(ac, av, "cr:d:o:pLCROPhxz:")) != EOF) {
		switch (i) {
		case 'c':
			color = 1;
			break;
		case 'r':
			parse_2dopt(optarg, &w, &h, 'r', 1);
			res_given = 1;
			break;
		case 'd':
			parse_2dopt(optarg, &xres, &yres, 'd', 1);
			break;
		case 'o':
			parse_2dopt(optarg, &xo, &yo, 'o', 0);
			break;
		case 'p':
			w = 640;
			h = 880;
			xres = yres = 75;
			align = 1;
			res_given = 1;
			break;
		case 'L':
			align = 0;
			break;
		case 'C':
			align = 1;
			break;
		case 'R':
			align = 2;
			break;
		case 'P':
			prerel = 1;
			break;
		case 'O':
			lampoff = 1;
			break;
		case 'x':
			xper = 1;
			break;
		case 'z':
			sscanf(optarg, "%d,%d,%d", &red_yoffset,
			       &green_yoffset, &blue_yoffset);
			max_yoffset = red_yoffset;
			if (green_yoffset > max_yoffset)
				max_yoffset = green_yoffset;
			if (blue_yoffset > max_yoffset)
				max_yoffset = blue_yoffset;
			break;
		case 'h':
		case '?':
			usage(av[0], i != 'h');
		}
	}
	outn = (optind < ac)? av[optind]: NULL;
	if (ac > optind + 1)
		usage(av[0], 1);
	if (!res_given && !lampoff)
		usage(av[0], 1);

	fd = open("/dev/usbscanner", O_RDWR);
	if (fd < 0) {
		perror("/dev/usbscanner");
		exit(1);
	}

	if (!res_given) {
		lamp(0);
		exit(0);
	}

	if (prerel) {
		align = 0;
		w *= xres / 75;
		h *= yres / 75;
		xo *= 600 / 75;
		yo *= 600 / 75;
		xo += (2607 - 640 * 8 / 2);
	}

	xs = 600 / xres;
	if (xs > 8 || xs * xres != 600 || !ISPOW2(xs)) {
		fprintf(stderr, "bad X resolution: "
			"must be 75, 150, 300 or 600\n");
		exit(1);
	}

	ydpi = 300;
	if (yres > 300)
		ydpi = 600;
	else if (yres <= 150 && color)
		ydpi = 150;
	ys = ydpi / yres;
	if (ys > 8 || ys * yres != ydpi || !ISPOW2(ys)) {
		fprintf(stderr, "bad Y resolution: "
			"must be 75, 150, 300 or 600\n");
		exit(1);
	}

	switch (align) {
	case 1:
		xo += 2607 - w * xs / 2;
		break;
	case 2:
		xo += 5214 - w * xs;
		break;
	}
	if (xo < -150 || xo + w * xs > 5400
	    || yo < -400 || yo + h * ys > 7200) {
		fprintf(stderr, "error: image corners out of range\n");
		exit(1);
	}
	xo += 152;

	umaxinit();
	umaxinit2();
	move(196, 0);
	find_zero();
	move(scanner_yorg - 232 - scanner_ypos, 1);
	get_caldata(color);

	if (outn == NULL) {
		f = stdout;
	} else {
		f = fopen(outn, "w");
		if (f == 0) {
			perror(outn);
			exit(1);
		}
	}

	if (color)
		yo -= 9;
	yd = scanner_yorg + yo - scanner_ypos;
	if (yd < 0) {
		move(yd, 1);
		yd = scanner_yorg + yo - scanner_ypos;
	}
	if (yd > 300) {
		move((yd - 20) / 2, 0);
		yd = scanner_yorg + yo - scanner_ypos;
	}
	yd /= 600 / ydpi;
	doscan((outn? outn: "(stdout)"), f, color, w, h, xs, ys, xo, yd, ydpi);
	if (outn != NULL)
		fclose(f);

	if (lampoff)
		opc1[14] = 0x90;
	restore();

	exit(0);
}
